import math
import os
import time


class BasicData:
    """
    Class for storing and tracking the basic training data in real time.\n
    Args:
        hr_data_path (str):
            path to the file that contains HR data
        gps_data_path (str):
            path to the file that contains GPS data
    """
    def __init__(self, hr_data_path: str, gps_data_path: str) -> None:
        """
        Initialization method for BasicData class.\n
        Args:
            hr_data_path (str):
                path to the file that contains HR data
            gps_data_path (str):
                path to the file that contains GPS data
        """
        self.hr_data_path = hr_data_path
        self.gps_data_path = gps_data_path
        self.current_speed = None
        self.current_heart_rate = None
        self.previous_gps = None
        self.current_gps = None
        self.distance = 0.0

        with open(self.gps_data_path, 'a') as f:
            f.truncate(0)
        with open(self.hr_data_path, 'a') as f:
            f.truncate(0)

    def read_current_hr(self) -> None:
        """
        Reading the current HR from the file.
        """
        if not os.path.exists(self.hr_data_path):
            self.current_heart_rate = None
            return

        with open(self.hr_data_path, 'rb') as f:
            try:
                f.seek(-2, os.SEEK_END)
                while f.read(1) != b'\n':
                    f.seek(-2, os.SEEK_CUR)
            except OSError:
                f.seek(0)
            try:
                hr = int(f.readline().decode().rstrip())
                self.current_heart_rate = str(hr)
            except ValueError:
                self.current_heart_rate = None

    def read_current_gps(self) -> None:
        """
        Reading the current GPS position from the file.
        """
        if not os.path.exists(self.gps_data_path):
            self.current_speed = None
            return

        with open(self.gps_data_path, 'rb') as f:
            try:
                f.seek(-2, os.SEEK_END)
                while f.read(1) != b'\n':
                    f.seek(-2, os.SEEK_CUR)
            except OSError:
                f.seek(0)
            try:
                self.previous_gps = self.current_gps

                # Reading the last GPS data.
                gps = f.readline().decode().rstrip().split(';')
                self.current_gps = (
                    float(gps[1]), float(gps[0]), float(gps[2]), float(gps[3])
                )
            except IndexError:
                self.previous_gps = None
                self.current_gps = None

    def calculate_speed(self) -> None:
        """
        Calculating the speed between the previous and the current trackpoint.
        """
        if not self.previous_gps or not self.current_gps:
            self.current_speed = None
            return

        # If the last measure is older than 1.5 seconds,
        # it is considered deprecated.
        if time.time() - self.current_gps[3] > 1.5:
            self.current_speed = None
            return

        # Calculating the distance.
        R = 6373.0
        lat1 = math.radians(self.previous_gps[0])
        lat2 = math.radians(self.current_gps[0])
        lon1 = math.radians(self.previous_gps[1])
        lon2 = math.radians(self.current_gps[1])
        dlon = lon2 - lon1
        dlat = lat2 - lat1
        a = (
            math.sin(dlat / 2)**2 +
            math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2)**2
        )
        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))

        self.distance = 1000 * R * c
        t = self.current_gps[3] - self.previous_gps[3]
        if t:
            self.current_speed = 3.6 * self.distance / t



import json


class RouteReader:
    def __init__(self, route_path):
        """
        Class for reading Json route generated by BikeRouteGen.
        Args:
            route_path: address where the route is saved.
        """
        self.route_path = route_path

    def read(self):
        """
        Read the route from the file.
        Returns:
            route: route dictionary
        """
        route = None
        with open(self.route_path) as f:
            json_route = json.load(f)
            route = self.merge_dicts([*json_route["sub_routes"]])
            route["route_render"] = json_route["route_render"]

        return route

    def merge_dicts(self, dict_list):
        """
        Helper method that merges dictionaries.
        Args:
            dict_list:

        Returns:
            merged_dict: merged dictionary
        """
        merged_dict = {}
        for key in dict_list[0].keys():  # Assuming all dictionaries have the same keys
            if isinstance(dict_list[0][key], list):
                merged_dict[key] = [item for d in dict_list for item in d[key]]
            elif isinstance(dict_list[0][key], dict):
                # Recursive call to handle nested dictionaries
                nested_dicts = [d[key] for d in dict_list]
                merged_dict[key] = self.merge_dicts(nested_dicts)
            else:
                # For non-list, non-dict types, you can choose how to handle them.
                # Here, I'm taking the value from the first dictionary for simplicity.
                merged_dict[key] = dict_list[0][key]
        return merged_dict

    def get_polyline(self, route):
        """
        Helper method that returns the polyline of the route.
        Args:
            route: route generated by BikeRouteGen and processed by read() method.

        Returns:
            polyline: list of tuples with the latitude and longitude of the route.
        """
        polyline = []
        for point in route["route_render"]:
            polyline.append((point["lat"], point["lon"]))
        return polyline



import http.server
import os
from urllib.parse import urlparse


class CustomHandler(http.server.SimpleHTTPRequestHandler):
    """Custom handler to serve the html files from the dist folder instead of the root folder.
    The handler helps us avoid changing the current directory of the Python script."""

    def translate_path(self, path):
        """
        Method that allows the server to serve the html files from the dist folder instead of the root folder.
        Args:
            path: url path

        Returns:

        """
        parsed_path = urlparse(path).path
        current_file_directory = os.path.dirname(os.path.abspath(__file__))
        custom_folder = os.path.join(current_file_directory, 'map', 'dist')
        translated_path = os.path.join(custom_folder, parsed_path.lstrip('/'))
        print(f"Translated path: {translated_path}")  # Debugging line to print the translated path
        return translated_path



import time
from threading import Thread

from ast_monitor.digital_twin import DigitalTwin
from ast_monitor.write_log import WriteLog


class IntervalTraining:
    """
    Initialization method of the IntervalTraining class.\n

    Args:
    ----
    training: dict
        an interval training
    tick_time: int
        tick time in milliseconds

    Returns:
    -------
    None
    """

    def __init__(self, training: dict, basic_data) -> None:
        """
        Initialization method of the IntervalTraining class.\n

        Args:
        ----
        training: dict
            an interval training
        tick_time: int
            tick time in milliseconds

        Returns:
        -------
        None
        """
        self.file = training['file']
        self.name = training['interval'][0]['name']
        self.sport = training['interval'][0]['sport']
        self.info = training['interval'][0]['info']
        self.speed_duration = training['interval'][0]['speed_duration']
        self.recovery_duration = training['interval'][0]['recovery_duration']
        self.speed_heart_rate = training['interval'][0]['speed_heart_rate']
        self.recovery_heart_rate = (
            training['interval'][0]['recovery_heart_rate']
        )
        self.repetitions = training['interval'][0]['repetitions']
        self.type = training['interval'][0]['type']
        self.speed_notification = False
        self.recovery_notification = False
        self.phase_start = None
        self.abort_training = False
        self.basic_data = basic_data
        name = self.file.split('.')[0]
        self.log = f'../logs/{name}.log'

    def __eq__(self, __o: object) -> bool:
        """
        Method for comparing two IntervalTraining objects.\n

        Args:
        ----
        __o: IntervalTraining
            an interval training to be compared to

        Returns:
        -------
        bool
        """
        return (
                self.name == __o.name and
                self.sport == __o.sport and
                self.info == __o.info and
                self.speed_duration == __o.speed_duration and
                self.recovery_duration == __o.recovery_duration and
                self.speed_heart_rate == __o.speed_heart_rate and
                self.recovery_heart_rate == __o.recovery_heart_rate and
                self.repetitions == __o.repetitions and
                self.type == __o.type
        )

    def start(self, write_log: bool = False) -> None:
        """
        Starting an interval training.\n

        Args:
        ----
        write_log: bool
            writing log if True

        Returns:
        -------
        None
        """
        # Writing the header of the training.
        if write_log:
            WriteLog.write_interval_training_header(self.log)

        for i in range(self.repetitions):
            self.start_speed_phase(i, write_log=True)
            if self.abort_training:
                break
            self.start_recovery_phase(i, write_log=True)
            if self.abort_training:
                break
        self.abort_training = True

    def start_speed_phase(self, interval: int, write_log: bool) -> None:
        """
        Starting a speed phase of an interval.\n

        Args:
        ----
        interval: int
            index of the interval
        write_log: bool
            writing log if True

        Returns:
        -------
        None
        """
        self.current_interval = interval, 'speed'
        self.speed_notification = True
        self.phase_start = time.time()
        self.time = time.time() - self.phase_start

        self.digital_twin = DigitalTwin(
            self.speed_heart_rate,
            self.speed_duration,
            tick_time=1000,
            basic_data=self.basic_data
        )
        self.digital_twin_thread = Thread(
            target=self.digital_twin.predict_heart_rate
        )
        self.digital_twin_thread.start()

        # Execution of a speed phase.
        while (
                not self.abort_training and
                self.time < 60 * self.speed_duration
        ):
            self.time = time.time() - self.phase_start
            if write_log:
                WriteLog.write_interval_training_trackpoint(
                    self.log,
                    self.digital_twin
                )
            time.sleep(0.5)

    def start_recovery_phase(self, interval: int, write_log: bool) -> None:
        """
        Starting a recovery phase of an interval.\n

        Args:
        ----
        interval: int
            index of the interval
        write_log: bool
            writing log if True

        Returns:
        -------
        None
        """
        self.current_interval = interval, 'recovery'
        self.recovery_notification = True
        self.phase_start = time.time()
        self.time = time.time() - self.phase_start

        self.digital_twin = DigitalTwin(
            self.recovery_heart_rate,
            self.recovery_duration,
            tick_time=1000,
            basic_data=self.basic_data
        )
        self.digital_twin_thread = Thread(
            target=self.digital_twin.predict_heart_rate
        )
        self.digital_twin_thread.start()

        # Execution of a recovery phase.
        while (
                not self.abort_training and
                self.time < 60 * self.recovery_duration
        ):
            self.time = time.time() - self.phase_start
            if write_log:
                WriteLog.write_interval_training_trackpoint(
                    self.log,
                    self.digital_twin
                )
            time.sleep(0.5)



from datetime import datetime
import time


class WriteLog:
    """
    Class for writing log.
    """
    @staticmethod
    def write_interval_training_header(file: str):
        """
        Method for writing interval training header.\n
        Args:
            file: str
                path to the log file
        """
        time_s = datetime.utcnow().isoformat()
        with open(file, 'a') as log:
            log.write(f'\n[{time_s}] {file}\n')
            log.write('Log parameters: HR_prop, HR_avg, HR_pred, t\n')

    @staticmethod
    def write_interval_training_trackpoint(file: str, digital_twin):
        """
        Method for writing interval training trackpoint.\n
        Args:
            file: str
                path to the log file
            digital_twin: DigitalTwin
                the Digital Twin
        """
        time_s = datetime.utcnow().isoformat()
        predicted_hr = '-'
        proposed_hr = str(int(digital_twin.proposed_heart_rate))
        average_hr = '-'
        phase_time = str(round(time.time() - digital_twin.start_time, 3))

        if hasattr(digital_twin, 'predicted_heart_rate'):
            predicted_hr = str(int(digital_twin.predicted_heart_rate))
        if digital_twin.average_heart_rate:
            average_hr = str(int(digital_twin.average_heart_rate))

        with open(file, 'a') as log:
            log.write(
                f'[{time_s}]' + ' ' +
                proposed_hr + ' ' +
                average_hr + ' ' +
                predicted_hr + ' ' +
                phase_time + '\n'
            )



from ast_monitor.basic_data import BasicData
from ast_monitor.digital_twin import DigitalTwin
from ast_monitor.goals_processor import GoalsProcessor
from ast_monitor.gps_sensor import GpsSensor
from ast_monitor.hr_sensor import HrSensor
from ast_monitor.interval_training import IntervalTraining
from ast_monitor.mainwindow import Ui_MainWindow
from ast_monitor.model import AST
from ast_monitor.route_reader import RouteReader
from ast_monitor.simulation import Simulation
from ast_monitor.training_session import TrainingSession
from ast_monitor.write_log import WriteLog

__all__ = [
    BasicData,
    DigitalTwin,
    GpsSensor,
    HrSensor,
    IntervalTraining,
    Ui_MainWindow,
    AST,
    GoalsProcessor,
    RouteReader,
    Simulation,
    TrainingSession,
    WriteLog
]

__project__ = 'ast_monitor'
__version__ = '0.4.3'



import time


class TrainingSession:
    """
    Class for tracking training sessions.
    """
    def __init__(self) -> None:
        self.start_time = time.time()
        self.time = 0.0
        self.distance = 0.0
        self.previous_elevation = None
        self.ascent = 0.0

    def calculate_time(self) -> None:
        """
        Method for calculating the current time of the session.
        """
        self.time = time.time() - self.start_time

    def add_distance(self, distance: float) -> None:
        """
        Adding to the total distance.\n
        Args:
            distance (float):
                distance in meters
        """
        self.distance += distance

    def add_ascent(self, current_elevation: float) -> None:
        """
        Adding to the total ascent.\n
        Args:
            current_elevation (float):
                current elevation in meters
        """
        if not self.previous_elevation:
            self.previous_elevation = current_elevation
            return

        diff = current_elevation - self.previous_elevation
        self.previous_elevation = current_elevation
        if diff > 0:
            self.ascent += diff



import math


class GoalsProcessor:
    def __init__(self, route):
        """Class for measuring the progress of the user on the route."""
        self.position = None
        self.total_ascent = route["evaluation"]["total_ascent"]
        self.total_distance = route["evaluation"]["total_distance"]
        self.ascents = route["ascents"]
        self.distances = route["distances"]
        self.current_distance = 0
        self.current_ascent = 0
        self.progress = 0
        self.distance_to_go = sum(self.distances)
        self.ascent_to_go = sum(self.ascents)
        self.index_closest = None
        self.route = route
        filteredIntersections = [sublist[1] for sublist in route["nodes"]]

        # Filter the list of dictionaries based on the IDs and remove duplicates
        filtered_list = []
        last_added_id = None

        for obj in route['route_render']:
            # Check if the id is in filteredIntersections
            if obj['id'] in filteredIntersections:
                # Check if the id is the same as the last added id
                if obj['id'] != last_added_id:
                    # Add the object to the filtered list
                    filtered_list.append(obj)

                    # Update the last added id
                    last_added_id = obj['id']

        # Now filtered_list contains the filtered and unique dictionaries
        self.filtered_list = filtered_list
        a = 100

    def add_position(self, position):
        """Add the current position of the user."""
        self.position = position[0:2]
        self.get_progress(self.position)

    def get_progress(self, position):
        """Calculate the progress of the user on the route, (%) of route, Ascent, Distance."""
        current_node = self.find_closest_object(position)
        distance = self.haversine_distance(position, (current_node['lat'], current_node['lon']))
        if distance < 0.05:
            self.current_distance += sum(self.distances[0:self.index_closest + 1])
            self.current_ascent += sum(self.ascents[0:self.index_closest + 1])
            # remove the summed indexes
            self.distances = self.distances[self.index_closest + 1:]
            self.ascents = self.ascents[self.index_closest + 1:]
            self.filtered_list = self.filtered_list[self.index_closest + 1:]
            self.index_closest = None
            self.distance_to_go = self.total_distance - self.current_distance
            self.ascent_to_go = self.total_ascent - self.current_ascent
            self.progress = self.current_distance / self.total_distance

    def find_closest_object(self, position):
        """Finds the closest object in the list to the given point."""
        smallest_diff = float('inf')
        closest_object = None

        index = 0
        # Loop through each object in the list
        for obj in self.filtered_list:
            lat_diff = math.fabs(obj['lat'] - position[0])
            lon_diff = math.fabs(obj['lon'] - position[1])

            # Calculate the total difference for both latitude and longitude
            total_diff = lat_diff + lon_diff

            # Update smallest_diff and closest_object if this object is closer to the given point
            if total_diff < smallest_diff:
                smallest_diff = total_diff
                closest_object = obj
                self.index_closest = index
            index += 1

        return closest_object  # Return the closest object

    def haversine_distance(self, coord1, coord2):
        """
        Calculate the Haversine distance between two points on the Earth specified by latitude/longitude.
        """

        # Radius of Earth in kilometers
        R = 6371.0

        # Convert latitude and longitude from degrees to radians
        lat1, lon1 = math.radians(coord1[0]), math.radians(coord1[1])
        lat2, lon2 = math.radians(coord2[0]), math.radians(coord2[1])

        # Compute differences in latitude and longitude
        dlat = lat2 - lat1
        dlon = lon2 - lon1

        # Haversine formula
        a = math.sin(dlat / 2) ** 2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2) ** 2
        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))

        distance = R * c

        return distance
